from ..block import Block
from ..abstract import ReportInterface
from typing import Dict


class Section(Block):
    """
    Class representing a section in the report

    """

    def __init__(self,
                 section_title: str):
        """

        Args:
            section_title (str): The title of the section
        """
        self.section_title = section_title
        self.content: Dict[str, Block] = {}

    def set_content(self, content: Dict[str, Block]):
        """
        Set the content of the section
        Args:
            content (dict): A dictionary of the content of the section, the keys are the titles of the blocks and the values are the blocks themselves

        Returns:

        """
        self.check_content(content)
        self.content = content

    def add_content(self, content: Dict[str, Block]):
        """
        Add content to the section
        Args:
            content (dict): A dictionary of the content of the section, the keys are the titles of the blocks and the values are the blocks themselves

        Returns:

        """
        self.check_content(content)
        self.content.update(content)

    @staticmethod
    def check_content(content: Dict[str, Block]):
        assert type(content) is dict

    def _as_html(self, report: ReportInterface):
        html = f"<div><h1 class=\"section_title editable\">{self.section_title}</h1></div>"
        for item in self.content.values():
            html += item._as_html(report)

        return html


class Subsection(Section):
    def _as_html(self, report: ReportInterface):
        html = f"<div><h2 class=\"section_title editable\">{self.section_title}</h2></div>"
        for item in self.content.values():
            html += item._as_html(report)

        return html



"""
Create html reports from python using simple objects

See module block for documentation on the blocks that can be added to reports.

Classes:
    Report
"""
from pathlib import Path
from datetime import datetime
from jinja2 import Template
from typing import Dict, Optional


from .block import Block
from .abstract import ReportInterface


class Report(ReportInterface):
    """
    Class representing a report
    """

    def __init__(self,
                 target_folder: Path = None,
                 template: Path = None,
                 filename: str = 'report',
                 template_content: Optional[Dict] = None,
                 ):
        """
        Create a report object
        Args:
            target_folder (Path): The folder where the report will be saved
            template (Path): The path to the template to be used for the report (Optional)
            filename: (str): The filename of the report (without suffix)
            template_content (dict): The content to be used for populating the template with the Jinja2 engine
        """

        if template_content is None:
            template_content = dict()
        self.template_content = template_content
        # To avoid long strings in documentation:
        if target_folder is None:
            raise AttributeError("Target folder is required")
        else:
            self.target_folder = target_folder

        if not self.target_folder.exists():
            self.target_folder.mkdir(parents=True)

        if template is None:
            self.template = Path(__file__).parent.joinpath('templates/report_template_numerous.html')
        else:
            self.template = template

        if not filename.endswith('.html'):
            filename += '.html'
        self.filename = filename

        self.report_header_info = None
        self.blocks = {}
        self.figure_number = 0
        self.table_number = 0

    def add_header_info(
            self,
            header: str,
            title: str,
            sub_title: str,
            sub_sub_title: str,
            footer_title: str="",
            footer_content: str=""
    ):
        """
        Adding information to be used for populating html template

        Args:
            header (str): String with header

            title (str): String with title

            sub_title (str):  String with sub title

            sub_sub_title (str): String with sub sub title

        """

        self.report_header_info = dict(
            report_header=header,
            report_type_title=header,
            report_title=title,
            report_sub_title=sub_title,
            report_sub_sub_title=sub_sub_title,
            report_date="{: %d-%m-%Y}".format(datetime.now()),
            footer_title=footer_title,
            footer_content=footer_content
        )

    def set_blocks(self, blocks: dict):
        """
        Overriding blocks to the report.

        Args:
            blocks (dict): Content of block. Values in the key can either be an instance of the class Tabs or Section.

        """
        self._check_blocks(blocks)
        self.blocks = blocks

    def add_blocks(self, blocks: dict):
        """
        Appending blocks to the report.

        Args:
            blocks (dict): Content of block. Values in the key can either be an instance of the class Tabs or Section.

        """
        self._check_blocks(blocks)
        self.blocks.update(blocks)

    def _check_blocks(self, blocks: dict):
        assert type(blocks) == dict, "blocks should be a dictionary"
        for instance in blocks.values():
            assert isinstance(instance, Block), \
                "values in blocks should be Section or Tabs Class instances"

    def generate_html(self):
        """
        Writing objects and header info into html template

        Returns: String html

        """
        assert self.report_header_info is not None, "Please fill out header info"

        html = ""
        for content in self.blocks.values():
            html += content._as_html(self)

        with open(self.template, 'r') as tf:
            template_str = tf.read()

        t = Template(template_str)

        self.template_content.update(self.report_header_info)

        rendered = t.render(report_content=html, **self.template_content).encode('utf-8')

        return rendered

    def save(self, images_as_bitmaps=False):
        """
        Args:
            images_as_bitmaps: Boolean to set if images are exported as bitmaps (PNG). False by default.

        Returns:

        """
        self.images_as_bitmaps = images_as_bitmaps
        """Saving the report to the target folder as html file"""
        with open(self.target_folder.joinpath(self.filename), 'wb') as outfile:
            outfile.write(self.generate_html())



from plotly import graph_objects as go
from plotly.basedatatypes import BaseTraceType
from typing import Union, Optional
from pathlib import Path

from ..figure import Figure
from ...abstract import ReportInterface


def wrap_div(html, class_=None):
    if class_:
        return f"<div class='{class_}'>\n{html}\n</div>\n"
    else:
        return f'<div>\n{html}\n</div>\n'


class GoFigure(Figure):
    """
    Class representing a plotly go figure in the report
    """
    def __init__(
            self,
            figure_data: Union[list[BaseTraceType], dict, go.Figure],
            caption: str = "",
            notes: Optional[list[str]] = None,
            post_script: str = ""
    ):
        """
        The constructor of the GoFigure class
        Args:
            figure_data (dict|go.Figure): The figure data or the go.Figure object
            caption(str): The caption text
            notes (List[str]): The notes to be added to the caption
            post_script: javascript code to be executed after the figure is rendered
        """
        super(GoFigure, self).__init__(caption, notes)
        self.figure_data = figure_data
        self.post_script = post_script
        self.figure_obj = go.Figure(self.figure_data)
        self.image_name: Optional[Path] = None

    def _as_html_figure_content(self, report: ReportInterface):

        if report.images_as_bitmaps:
            self.image_name = report.target_folder / Path(self.id + ".png")
            self.figure_obj.write_image(
                file=report.target_folder / Path(self.id + ".png"),
                format="png"
            )
            html = wrap_div(
                html=f"<img src={self.image_name.name} alt=Figure {report.figure_number}",
                class_="figure_div",
            )
        else:

            html = wrap_div(
                self.figure_obj.to_html(
                    config={"displayModeBar": True},
                    include_plotlyjs=False,
                    full_html=False,
                    post_script=self.post_script
                ),
                class_='figure_div'
            )

        return html



import pandas as pd

from ..table import Table
from typing import List, Optional, AnyStr


class DataFrameTable(Table):
    """
    Create a dataframe (pandas) table, which may be formatted using html classes

    Args:
        table_df: the pandas dataframe table
        caption: a caption applied to the top of the html table
        notes: notes to added
        show_index: True if the index column should be shown, false if not (default)
        classes: the html classes to format the table with

    """

    def __init__(
            self,
            table_df: pd.DataFrame,
            caption: str = "",
            notes: Optional[List[str]] = None,
            show_index: bool = False,
            classes: Optional[List | AnyStr] = None
    ):
        """

        Args:
            table_df (pd.DataFrame): The pandas dataframe to be used for the table
            caption (str): The caption text
            notes (List[str]): The notes to be added to the caption
            show_index (bool): Whether to show the index column
            classes (List[str]): The html classes to apply to the table
        """
        if not notes:
            notes = []
        super(DataFrameTable, self).__init__(caption, notes)

        self.table_df = table_df
        self.show_index = show_index
        self.classes = classes

    def as_html_figure_content(self):
        return self.table_df.to_html(classes=self.classes, index=self.show_index, render_links=True, escape=False)



from abc import ABC, abstractmethod


class Block(ABC):
    """
    Abstract class representing a block in the report
    This is not meant to be used directly but to be inherited by specific block classes
    """
    @abstractmethod
    def _as_html(self, *args):
        ...



import uuid

from ..block import Block
from typing import Optional
from ..caption import Caption
from ..abstract import ReportInterface


class Figure(Block):
    """
    Class representing a figure in the report
    This is not meant to be used directly but to be inherited by specific figure classes
    """
    def __init__(self, caption: str, notes: Optional[list[str]] = None):
        """

        Args:
            caption (str): The caption text
            notes (List[str]): The notes to be added to the caption
        """
        self.caption = Caption(caption=caption, notes=notes, type="Figure")
        self.id = str(uuid.uuid4())

    def _as_html_figure_content(self, report: ReportInterface):
        return ""

    def _as_html(self, report: ReportInterface):
        report.figure_number += 1
        return f"{self._as_html_figure_content(report) + self.caption.caption_as_html(report.figure_number)}"



r"""
======================
Quick Start Example
======================

Simple example that creates a report with a figure and a table and adds them to a section.
"""

from pathlib import Path
import os
from numerous.html_report_generator import Report, Section, GoFigure, DataFrameTable
import pandas as pd

# Define output folder and html file
folder = Path('./output')
filename = 'my_first_report'
file = folder.joinpath(filename+'.html')

#Remove previous report file if exists
if os.path.exists(file):
    os.remove(file)

# Create report
report = Report(target_folder=folder, filename=filename)

# Add info for the header and title page
report.add_header_info(header='My Organization',
                       title='My First Report',
                       sub_title='but not the last',
                       sub_sub_title='at all',
                       footer_title='My first footer',
                       footer_content='This is the end!'
                       )

#Add a section
section = Section(section_title="Section 1")

#create a figure
figure = GoFigure({
        "data": [{"type": "bar",
                  "x": [1, 2, 3],
                  "y": [1, 3, 2]}],
        "layout": {"title": {"text": "A Figure Specified By Python Dictionary"}}
    }, caption="Test figure", notes=["Please notice x", "Please notice y"])

#create a table
table = DataFrameTable(pd.DataFrame({
    'a': [1, 2, 3],
    'b': [4, 5, 6]
}), caption="This is a test table")

#Add the figure and table to the section
section.add_content({
    'figure': figure,
    'table': table
})

#Add the section to the report
report.add_blocks({
    'section': section
})

# Save the report - creates the html output file
report.save()



from typing import Dict
from abc import ABC, abstractmethod
from pathlib import Path


class ReportInterface(ABC):
    figure_number: int
    table_number: int
    target_folder: Path
    images_as_bitmaps: bool = False

    @abstractmethod
    def add_header_info(
            self,
            header: str,
            title: str,
            sub_title: str,
            sub_sub_title: str,
            footer_title: str,
            footer_content: str,
    ):
        pass

    @abstractmethod
    def set_blocks(
            self,
            blocks: Dict,
            ):
        pass

    @abstractmethod
    def add_blocks(
            self,
            blocks: Dict,
    ):
        pass

    @abstractmethod
    def save(self):
        pass

    @abstractmethod
    def generate_html(self):
        pass









"""
An extensible python package that enables you to easily create great looking modular html reports

When you have created your data science script and have all the nice results you want to package this up in a nice way.
To allow you share with your peers.
This package helps you create interactive and nice looking reports, yet available in a single static html file.
"""
from .report import Report
from numerous.html_report_generator.components.plotly.gofigure import GoFigure
from .components.div import Div
from .components.tabs import Tabs
from .components.section import Section, Subsection
from .components.cards import Card
from numerous.html_report_generator.components.pandas.dataframetable import DataFrameTable


from typing import Optional


class Caption:
    """
    Class representing a caption for a figure or table
    """
    def __init__(self, caption: str, notes: Optional[list[str]] = None, type: str = "Figure"):
        """

        Args:
            caption: The caption text. An empty caption disables the caption and numbering.
            notes:  Notes to be added to the caption
            type: A string indicating the type of caption, e.g. "Figure" or "Table"
        """
        self.caption = caption
        if notes:
            self.notes = notes
        else:
            self.notes = []
        self.type = type

    def caption_as_html(self, number: int):
        """
        Returns the caption as html

        """
        notes_str = ''.join([f'<div class="note"> <i>Note: {n} </i></div>' for n in self.notes])

        if self.caption:
            return f'<div class="caption-section"><div class="caption"><b>{self.type} {number}:</b> {self.caption}</div>{notes_str}</div>'
        else:
            return f'<div class="caption-section">{notes_str}</div>'



from numerous.html_report_generator.block import Block


class Div(Block):
    """
        Class representing a div in the report
    """
    def __init__(self, html:str, **kwargs):
        """

        Args:
            html: The html content of the div
            **kwargs:
        """
        self.html = html
        self.kwargs = kwargs

    def _as_html(self, *args):
        modifiers=[]

        for k, v in self.kwargs.items():
            modifiers.append(f'{k}="{v}"')

        return f"<div {' '.join(modifiers)}>{self.html}</div>"



from itertools import count
from numerous.html_report_generator.block import Block
from ..abstract import ReportInterface


class Tabs(Block):
    """
    Class representing a tabs object in the report

    Attributes:
        id (int): Unique id for each tabs object
        tabs (dict): Content of tabs. Each key will be used as name of the buttons.
            The values of the dict will be used as html content in the tab.
    """

    _ids = count(0)
    _unique_tab_id = count(0)

    def __init__(self):

        self.id = next(self._ids)
        self.tabs = {}

    def set_tabs(self, tabs: dict):
        """
        Set the content of the tabs
        Args:
            tabs (dict): A dictionary of the content of the tabs, the keys are the titles of the tabs and the values are the content of the tabs

        Returns:

        """
        self.check_tabs(tabs)
        self.tabs = tabs

    def add_tabs(self, tabs:dict):
        """
        Add content to the tabs
        Args:
            tabs (dict): A dictionary of the content of the tabs, the keys are the titles of the tabs and the values are the content of the tabs

        Returns:

        """
        self.check_tabs(tabs)
        self.tabs.update(tabs)

    def check_tabs(self, tabs):
        assert type(tabs) == dict


    def _as_html(self, report: ReportInterface):

        tab_script = """\n\n<script>
                function openTab_""" + str(self.id) + """(evt, tabname) {
                  var i, tabcontent, tablinks;
                  tabcontent = document.getElementById('""" + str(self.id) + """').getElementsByClassName("tabcontent");

                  for (i = 0; i < tabcontent.length; i++) {

                    tabcontent[i].style.display = "none";
                  }

                  tablinks = document.getElementById('""" + str(self.id) + """').getElementsByClassName("tablinks");
                  for (i = 0; i < tablinks.length; i++) {
                    tablinks[i].className = tablinks[i].className.replace(" active", "");
                  }
                  thistab = document.getElementById(tabname);
                  //thistab.style.width= '100%';
                  thistab.style.display = "block";
                  //window.dispatchEvent(new Event('resize'));
                  evt.currentTarget.className += " active";
                }
                var evt = document.createEvent("MouseEvents");
                evt.initMouseEvent("click", true, true, window, 1, 0, 0, 0, 0,
                    false, false, false, false, 0, null);

                document.getElementById('default_tab_""" + str(self.id) + """').dispatchEvent(evt);
                </script>\n\n"""
        tab_divs = []
        tab_buttons = []

        for label, content in self.tabs.items():

            unique_tab_id = next(self._unique_tab_id)
            html_context = content._as_html(report) if isinstance(content, Block) else content

            assert isinstance(html_context, str), "The content of the tabs must be a string or a Block object"

            tab_label = f"""
            <button 
                id=default_tab_{self.id} 
                class="tablinks" 
                onclick="openTab_{self.id}(event, 'tab_{unique_tab_id}')">{label}
            </button>"""

            tab_div = f"""
            <div id="tab_{unique_tab_id}" class="tabcontent" style="display: block;" width="100%">
                  {html_context}
            </div>"""

            tab_divs.append(tab_div)
            tab_buttons.append(tab_label)

        buttons = "".join(tab_buttons)
        tabs = "".join(tab_divs)
        html = f'<div id="{self.id}"><div class="tab">{buttons}</div>{tabs}</div>{tab_script}'

        return html



from ..block import Block
from typing import List
from ..caption import Caption
from ..abstract import ReportInterface


class Table(Block):
    """
    Class representing a table in the report
    This is not meant to be used directly but to be inherited by specific table classes
    """
    def __init__(
            self,
            caption: str,
            notes: List[str],
    ):
        self.caption = Caption(caption=caption, notes=notes, type="Table")

    # Overwrite this function in specific figure class
    def as_html_figure_content(self):
        return ""

    def _as_html(self, report: ReportInterface):
        report.table_number += 1
        return f"{self.as_html_figure_content()+self.caption.caption_as_html(report.table_number)}"




from typing import Optional, List, AnyStr
from .section import Section
from ..abstract import ReportInterface


class Card(Section):
    """
    Create a card that can be formatted with html classes, and which may have content

    Args:
        id: the id applied to the card
        classes: the html classes to apply to the card
    """
    def __init__(self, id: Optional[AnyStr] = None, classes: Optional[List | AnyStr] = None):
        """

        """
        super().__init__(section_title="")
        self.id = id
        self.classes = classes

    def _as_html(self, report: ReportInterface):
        html = f"<div class={self.classes} id={self.id}>"
        for item in self.content.values():
            html += item._as_html(report)
        html += '</div>'
        return html
